/***************************************************************************************
  *
  * dpcstkgr.c - DPC Stack Grower Kernel Module for Windows 2000/XP/2003 - Version 1.0
  *
  * Copyright (c)2004 VitoPlantamura.COM. All Rights Reserved.
  *
***************************************************************************************/

//===========
// Includes.
//===========

#include "..\Include\dpcstkgr.h"
#include "..\Include\IOCTLs.h"
#include "..\Include\detours.h"

//================
// Detours stuff.
//================

static PVOID		g_pfnKiRetireDpcListFN = NULL;
DETOUR_TRAMPOLINE_GLOBVAR( VOID Trampoline_KiRetireDpcListFN( VOID ), g_pfnKiRetireDpcListFN )
VOID Hooked_KiRetireDpcListFN( VOID );

//==================================
// Boilerplate KM device functions.
//==================================

NTSTATUS VPCKernelNtDispatchCreate( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
NTSTATUS VPCKernelNtDispatchClose( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
NTSTATUS VPCKernelNtDispatchIoctl( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );

//=================================================
// Private storage for us to squirrel things away.
//=================================================

// *** Device Object Global Pointer. ***

PDEVICE_OBJECT		g_pDeviceObject = NULL;

//==================================
// DriverEntry Function Definition.
//==================================

NTSTATUS DriverEntry( IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS        ntStatus;
	UNICODE_STRING  uszDriverString;
	UNICODE_STRING  uszDeviceString;

	PDEVICE_OBJECT    pDeviceObject;
	PDEVICE_EXTENSION extension;

	// Point uszDriverString at the driver name.

	RtlInitUnicodeString( & uszDriverString, L"\\Device\\" SYSDRIVER_NAME_WIDE );

	// Create and initialize device object.

	ntStatus = IoCreateDevice( DriverObject,
		sizeof( DEVICE_EXTENSION ),
		& uszDriverString,
		FILE_DEVICE_UNKNOWN,
		0,
		FALSE,
		& pDeviceObject );

	if ( ntStatus != STATUS_SUCCESS )
	{
		// Return to the Caller.

		return ntStatus;
	}

	// Assign extension variable.

	extension = pDeviceObject->DeviceExtension;

	// Point uszDeviceString at the device name.

	RtlInitUnicodeString( & uszDeviceString, L"\\DosDevices\\" SYSDRIVER_NAME_WIDE );

	// Create symbolic link to the user-visible name.

	ntStatus = IoCreateSymbolicLink( & uszDeviceString, & uszDriverString );

	if ( ntStatus != STATUS_SUCCESS )
	{
		// Delete device object if not successful.

		IoDeleteDevice( pDeviceObject );

		return ntStatus;
	}

	// Assign global pointer to the device object.

	g_pDeviceObject = pDeviceObject;

	// Load structure to point to IRP handlers.

	DriverObject->DriverUnload                           = NULL; // ### can't be unloaded ! ###
	DriverObject->MajorFunction[ IRP_MJ_CREATE ]         = VPCKernelNtDispatchCreate;
	DriverObject->MajorFunction[ IRP_MJ_CLOSE ]          = VPCKernelNtDispatchClose;
	DriverObject->MajorFunction[ IRP_MJ_DEVICE_CONTROL ] = VPCKernelNtDispatchIoctl;

	// Initialize WasCreated and its mutex.

	extension->WasCreated = 0;
	ExInitializeFastMutex( & extension->WasCreatedFastMutex );

	// Initialize the hook stuff.

	extension->dwNumberProcessors = (DWORD) * KeNumberProcessors;
	memset( & extension->vppdssState, 0, sizeof( extension->vppdssState ) );

	KeInitializeSpinLock( & extension->slStateSpinLock );
	extension->ulUseStateSpinLock = 1;
	extension->ulNumProcAllocated = 0;

	// Return Success.

	return STATUS_SUCCESS;
}

//================================================
// VPCKernelNtDispatchCreate Function Definition.
//================================================

NTSTATUS VPCKernelNtDispatchCreate( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
	PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
	ULONG             WasCreated;
	NTSTATUS          ntStatus;

	// Allow only one creation of the file object

	ExAcquireFastMutex( & extension->WasCreatedFastMutex );
	{
		WasCreated = extension->WasCreated;
		if ( extension->WasCreated == 0 )
			extension->WasCreated = 1;
	}
	ExReleaseFastMutex( & extension->WasCreatedFastMutex );

	if ( WasCreated != 0 )
		ntStatus = STATUS_UNSUCCESSFUL;
	else
		ntStatus = STATUS_SUCCESS;

	// Return to the caller

	Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return ntStatus;
}

//===============================================
// VPCKernelNtDispatchClose Function Definition.
//===============================================

NTSTATUS VPCKernelNtDispatchClose( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
	PDEVICE_EXTENSION		extension = DeviceObject->DeviceExtension;
	ULONG					bDoCleanUp = 0;

	// Remember that the client is closing the device

	ExAcquireFastMutex( & extension->WasCreatedFastMutex );

		if ( extension->WasCreated )
			bDoCleanUp = 1;
		extension->WasCreated = 0;

	ExReleaseFastMutex( & extension->WasCreatedFastMutex );

	// Check whether we have to do Clean-Up.

	if ( bDoCleanUp )
	{
		// ### clean-up code here ! ###
	}

	// Return to the caller

	if ( Irp )
	{
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;

		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	}

    return STATUS_SUCCESS;
}

//===============================================
// VPCKernelNtDispatchIoctl Function Definition.
//===============================================

NTSTATUS VPCKernelNtDispatchIoctl( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    NTSTATUS              ntStatus = STATUS_UNSUCCESSFUL;
    PIO_STACK_LOCATION    irpStack  = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION     extension = DeviceObject->DeviceExtension;

	DWORD*                pl;

	// IOCTL Processor Switch.

	switch( irpStack->Parameters.DeviceIoControl.IoControlCode )
	{
		//-----------------------------------------------------------------------------------

		case IOCTL_DPCSTKGRNT_GET_VERSION:
			{
				// Return the information required.

				if ( irpStack->Parameters.DeviceIoControl.OutputBufferLength >= sizeof( DWORD ) )
				{
					// Copy the information in the system buffer.

					pl = Irp->AssociatedIrp.SystemBuffer;
					*pl = ( DWORD ) _VPCDpcSG_DRIVER_VERSION_;

					// Everything went ok.

					ntStatus = STATUS_SUCCESS;
				}
			}
			break;

		//-----------------------------------------------------------------------------------

		case IOCTL_DPCSTKGRNT_INITIALIZE:
			{
				// === Hook the system function. ===

				if ( irpStack->Parameters.DeviceIoControl.InputBufferLength == INITIALIZE_INPUT_SIZE &&
					irpStack->Parameters.DeviceIoControl.OutputBufferLength == INITIALIZE_OUTPUT_SIZE )
				{
					INITIALIZE_INPUT			iiInput;
					INITIALIZE_OUTPUT			ioOutput;
					BOOLEAN						bWritProt = FALSE;
					PVOID						pvKiRetireDpcList;
					BOOLEAN						bHookRes;
					DELAY_TARGET_JUMP_INSERTION	dtjiJmpInsertInfo;
					INT							i, j;
					BOOLEAN						bAllocErr;
					PVOID						pv;
					DWORD						sz;

					memset( & dtjiJmpInsertInfo, 0, sizeof( dtjiJmpInsertInfo ) );

					// Initialize.

					iiInput = * (INITIALIZE_INPUT*) Irp->AssociatedIrp.SystemBuffer;
					memset( & ioOutput, 0, sizeof( ioOutput ) );

					ntStatus = STATUS_SUCCESS; // ---> We return Detailed Informations about errors in the Output Structure.

					// Hook.

					if ( extension->dwNumberProcessors <= MACRO_MAX_NUMBERPROCESSORS &&
						IsEnforceWriteProtectionSetTo0( & bWritProt ) == STATUS_SUCCESS &&
						bWritProt )
					{
						pvKiRetireDpcList = Guess_KiRetireDpcList_FnPtr ();
						if ( pvKiRetireDpcList )
						{
							ioOutput.dwKiRetireDpcListFnAddr = pvKiRetireDpcList;

							// disassemble.

							__try
							{
								g_pfnKiRetireDpcListFN = pvKiRetireDpcList;
								bHookRes = DetourFunctionWithTrampoline( (PBYTE) Trampoline_KiRetireDpcListFN, (PBYTE) Hooked_KiRetireDpcListFN, & dtjiJmpInsertInfo );
							}
							__except ( EXCEPTION_EXECUTE_HANDLER )
							{
								bHookRes = FALSE;
							}

							if ( bHookRes &&
								dtjiJmpInsertInfo.cbCode >= 5 )
							{
								// allocate memory.

								bAllocErr = FALSE;

								for( i=0; i<(int)extension->dwNumberProcessors; i ++ )
								{
									sz = iiInput.ulStackSizeInKBytes * 1024;
									pv = (PVOID) ExAllocatePool( NonPagedPool, sz );

									if ( pv )
									{
										// remember this ptr.
										extension->vppdssState[ i ].pvNewStack = pv;
										extension->vppdssState[ i ].dwNewStackPointer = (DWORD) pv + sz;
									}
									else
									{
										// free mem.
										for( j=0; j<i; j ++ )
											ExFreePool( extension->vppdssState[ j ].pvNewStack );

										// exit.
										bAllocErr = TRUE;
										break;
									}
								}

								if ( bAllocErr == FALSE )
								{
									// install hook.

									__asm
									{
										lea		edx, dtjiJmpInsertInfo

										mov		ebx, [ edx ]DELAY_TARGET_JUMP_INSERTION.pbDest
										sub		ebx, [ edx ]DELAY_TARGET_JUMP_INSERTION.pbCode
										sub		ebx, 5

										mov		eax, [ edx ]DELAY_TARGET_JUMP_INSERTION.pbCode

										cli									// =========================================================================================================
										mov		byte ptr[ eax ], 0xE9		// =  W A R N I N G                                                                                        =
										mov		dword ptr[ eax + 1 ], ebx	// =    there is a minimal possibility of a race condition on SMP.                                         =
										sti									// =    the correct implementation would require a spinlock and a serializing instruction, such as CPUID.  =
																			// =========================================================================================================
										add		eax, 5

										mov		ecx, [ edx ]DELAY_TARGET_JUMP_INSERTION.cbCode
										sub		ecx, 5

										push	edi
										mov		edi, eax
										mov		al, 0xCC
										rep		stosb
										pop		edi
									}

									// return success.

									ioOutput.bSucceeded = TRUE;
								}
							}
						}
					}

					// Copy the Header structure.

					* (INITIALIZE_OUTPUT*) Irp->AssociatedIrp.SystemBuffer = ioOutput;
				}
			}
			break;

		//-----------------------------------------------------------------------------------

        default:
            break;

		//-----------------------------------------------------------------------------------
	}

	// Set the Return Value.

    Irp->IoStatus.Status = ntStatus;
    
    // Set # of bytes to copy back to user-mode.

    if ( ntStatus == STATUS_SUCCESS )
        Irp->IoStatus.Information = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    else
        Irp->IoStatus.Information = 0;

	// Return to the Caller.

    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return ntStatus;
}

//=====================================================
// IsEnforceWriteProtectionSetTo0 Function Definition.
//=====================================================

NTSTATUS IsEnforceWriteProtectionSetTo0( OUT BOOLEAN* pbResult )
{
	NTSTATUS						nsRetVal = STATUS_UNSUCCESSFUL;
	UNICODE_STRING					usMemoryManagementKeyName;
	OBJECT_ATTRIBUTES				oaMemoryManagementKeyAttr;
	NTSTATUS						nsOpenMemoryManagementKeyRes;
	HANDLE							hMemoryManagementKeyHandle;
	PKEY_VALUE_FULL_INFORMATION		pkvfiEnforceWriteProtectionValueBuffer;
	ULONG							ulEnforceWriteProtectionValueBufferLen = 8 * 1024;
	UNICODE_STRING					usEnforceWriteProtectionValueName;
	NTSTATUS						nsReadEnforceWriteProtectionValueRes;

	* pbResult = FALSE;

	// Read from the Registry.

	RtlInitUnicodeString( & usMemoryManagementKeyName, L"\\REGISTRY\\MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Memory Management" );

	InitializeObjectAttributes( & oaMemoryManagementKeyAttr,
		& usMemoryManagementKeyName, 0, NULL, NULL );

	nsOpenMemoryManagementKeyRes = ZwOpenKey( & hMemoryManagementKeyHandle,
		KEY_READ, & oaMemoryManagementKeyAttr );

	if ( nsOpenMemoryManagementKeyRes == STATUS_SUCCESS && hMemoryManagementKeyHandle )
	{
		pkvfiEnforceWriteProtectionValueBuffer = (PKEY_VALUE_FULL_INFORMATION) ExAllocatePool( PagedPool, ulEnforceWriteProtectionValueBufferLen );

		if ( pkvfiEnforceWriteProtectionValueBuffer )
		{
			RtlInitUnicodeString( & usEnforceWriteProtectionValueName, L"EnforceWriteProtection" );

			nsReadEnforceWriteProtectionValueRes = ZwQueryValueKey( hMemoryManagementKeyHandle,
				& usEnforceWriteProtectionValueName, KeyValueFullInformation,
				pkvfiEnforceWriteProtectionValueBuffer, ulEnforceWriteProtectionValueBufferLen, & ulEnforceWriteProtectionValueBufferLen );

			if ( nsReadEnforceWriteProtectionValueRes == STATUS_SUCCESS &&
				ulEnforceWriteProtectionValueBufferLen >= sizeof( KEY_VALUE_FULL_INFORMATION ) )
			{
				if ( pkvfiEnforceWriteProtectionValueBuffer->Type == REG_DWORD &&
					pkvfiEnforceWriteProtectionValueBuffer->DataLength == sizeof( DWORD ) &&
					* (DWORD*) ( ( BYTE* ) pkvfiEnforceWriteProtectionValueBuffer + pkvfiEnforceWriteProtectionValueBuffer->DataOffset ) == 0 )
				{
					* pbResult = TRUE;
				}

				nsRetVal = STATUS_SUCCESS;
			}

			ExFreePool( pkvfiEnforceWriteProtectionValueBuffer );
		}

		ZwClose( hMemoryManagementKeyHandle );
	}

	// Return to the Caller.

	return nsRetVal;
}

//==================================================
// Guess_KiRetireDpcList_FnPtr Function Definition.
//==================================================

static PVOID Guess_KiRetireDpcList_FnPtr_IMPL( IN BYTE* pbPtr, IN BYTE bPattern00, IN DWORD dwPattern01, IN DWORD dwPattern02, IN INT iDir )
{
	VOID*		pvRetVal = NULL;
	ULONG		ulI;

	// search in memory.

	__try
	{
		for ( ulI = 0; ulI < MACRO_KIRETIREDPCLIST_FNSRCH_RANGE; ulI ++, pbPtr += iDir )
		{
			if ( * pbPtr == bPattern00 )
				if ( * (DWORD*) ( pbPtr + 1 ) == dwPattern01 &&
					* (DWORD*) ( pbPtr + 5 ) == dwPattern02 )
				{
					pvRetVal = (VOID*) pbPtr;
					break;
				}
		}
	}
	__except ( EXCEPTION_EXECUTE_HANDLER )
	{ }

	// return.

	return pvRetVal;
}

PVOID Guess_KiRetireDpcList_FnPtr( VOID )
{
	BYTE*		pbPtr00 = (BYTE*) & KeInitializeSpinLock;
	BYTE*		pbPtr01 = (BYTE*) & KiDeliverApc;
	BYTE*		pbPtr02 = (BYTE*) & KiCoprocessorError;
	BYTE*		pbPtr03 = (BYTE*) & KiDispatchInterrupt;

	VOID*		pvRetVal = NULL;

	// Search for the Beginning of the Function.

	pvRetVal = Guess_KiRetireDpcList_FnPtr_IMPL( // windows 2000 smp free.
		pbPtr00,
		MACRO_KIRETIREDPCLIST_FNSRCH_W2KSMPFREE_1STBYTE,
		MACRO_KIRETIREDPCLIST_FNSRCH_W2KSMPFREE_2NDDWORD,
		MACRO_KIRETIREDPCLIST_FNSRCH_W2KSMPFREE_3NDDWORD,
		-1 );

	if ( pvRetVal == NULL )
	{
		pvRetVal = Guess_KiRetireDpcList_FnPtr_IMPL( // windows XP smp free.
			pbPtr01,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPSMPFREE_1STBYTE,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPSMPFREE_2NDDWORD,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPSMPFREE_3NDDWORD,
			-1 );
	}

	if ( pvRetVal == NULL )
	{
		pvRetVal = Guess_KiRetireDpcList_FnPtr_IMPL( // windows 2000 up free.
			pbPtr02,
			MACRO_KIRETIREDPCLIST_FNSRCH_W2KUPFREE_1STBYTE,
			MACRO_KIRETIREDPCLIST_FNSRCH_W2KUPFREE_2NDDWORD,
			MACRO_KIRETIREDPCLIST_FNSRCH_W2KUPFREE_3NDDWORD,
			-1 );
	}

	if ( pvRetVal == NULL )
	{
		pvRetVal = Guess_KiRetireDpcList_FnPtr_IMPL( // windows XP up free.
			pbPtr03,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPUPFREE_1STBYTE,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPUPFREE_2NDDWORD,
			MACRO_KIRETIREDPCLIST_FNSRCH_WXPUPFREE_3NDDWORD,
			+1 );
	}

	return pvRetVal;
}

//===============================================
// Hooked_KiRetireDpcListFN Function Definition.
//===============================================

VOID __declspec( naked ) Hooked_KiRetireDpcListFN( VOID )
{
	__asm
	{
		// Save the state of the processor.

		push		0x00000000 // USED TO STORE THE NEW STACK POINTER.

#define MACRO_REGSTATE_SIZE			( 5 * 4 )
		push		eax // .
		push		ebx // . IF YOU ADD A NEW PUSH HERE, UPDATE THE "MACRO_REGSTATE_SIZE" CONSTANT.
		push		ecx // .
		push		edx // .
		push		esi // .

		// Get a pointer to the device extension.

		mov			ebx, g_pDeviceObject                      //  ...
		mov			ebx, [ ebx ]DEVICE_OBJECT.DeviceExtension // DEVICE EXTENSION

		// Acquire the SPINLOCK, if required.

		mov			esi, [ ebx ]DEVICE_EXTENSION.ulUseStateSpinLock
		or			esi, esi
		jz			_skip_sl

		push		eax
		push		ebx
		push		ecx
		push		edx
		push		esi
		push		edi

			lea		ecx, [ ebx ]DEVICE_EXTENSION.slStateSpinLock // .. fastcall ..
			call	dword ptr [KefAcquireSpinLockAtDpcLevel] // ON UP, DISPATCH_LEVEL WILL SUFFICE FOR SYNC.

		pop			edi
		pop			esi
		pop			edx
		pop			ecx
		pop			ebx
		pop			eax

_skip_sl:
		// Iterate in the list of stack pointers.

		lea			edx, [ ebx ]DEVICE_EXTENSION.vppdssState // STATE VECTOR PTR
		mov			ecx, [ ebx ]DEVICE_EXTENSION.dwNumberProcessors // NUM PROCS

		mov			eax, FS:[ 0x1C ] // PCR ADDRESS
_proc_loop:
		cmp			[ edx ]PERPROC_DPCSTACK_STATE.pvRefPcr, eax
		je			_state_found
		cmp			[ edx ]PERPROC_DPCSTACK_STATE.pvRefPcr, 0
		je			_slot_nf
		add			edx, SIZE PERPROC_DPCSTACK_STATE
		loop		_proc_loop

_slot_nf:
		// The stack pointer was not found.

		mov			eax, [ ebx ]DEVICE_EXTENSION.ulNumProcAllocated
		cmp			eax, [ ebx ]DEVICE_EXTENSION.dwNumberProcessors
		jae			_exit_from_fn // ## ## INTERNAL BUG ## ##: we have NOT found the processor and all the processors are allocated...

		// Assign the processor to the corresponding PERPROC_DPCSTACK_STATE structure.

		mov			eax, FS:[ 0x1C ] // PCR ADDRESS
		mov			[ edx ]PERPROC_DPCSTACK_STATE.pvRefPcr, eax

		inc			[ ebx ]DEVICE_EXTENSION.ulNumProcAllocated

		// Check whether we have to disable the SPINLOCK utilization.

		mov			eax, [ ebx ]DEVICE_EXTENSION.ulNumProcAllocated
		cmp			eax, [ ebx ]DEVICE_EXTENSION.dwNumberProcessors
		jb			_state_found

		mov			[ ebx ]DEVICE_EXTENSION.ulUseStateSpinLock, 0

_state_found:
		// Get the new stack pointer value.

		mov			ecx, esp
		add			ecx, MACRO_REGSTATE_SIZE + 4 // OLD STACK POINTER VALUE

		mov			eax, [ edx ]PERPROC_DPCSTACK_STATE.dwNewStackPointer
		sub			eax, 4
		mov			dword ptr[ eax ], ecx
		mov			dword ptr[ esp + MACRO_REGSTATE_SIZE ], eax

_exit_from_fn:

		// Release the SPINLOCK, if required.

		or			esi, esi
		jz			_skip_sl_2

		push		eax
		push		ebx
		push		ecx
		push		edx
		push		esi
		push		edi

			lea		ecx, [ ebx ]DEVICE_EXTENSION.slStateSpinLock // .. fastcall ..
			call	dword ptr [KefReleaseSpinLockFromDpcLevel]

		pop			edi
		pop			esi
		pop			edx
		pop			ecx
		pop			ebx
		pop			eax

_skip_sl_2:
		// Restore the state of the processor.

		pop			esi
		pop			edx
		pop			ecx
		pop			ebx
		pop			eax

		pop			esp

		// Call the original function.

		call		Trampoline_KiRetireDpcListFN

		// Restore the old SP.

		pop			esp

		// return.

		ret
	}
}
